import struct

_range = getattr(__builtins__, 'xrange', range)

class BinaryInput(object):
    def __init__(self, buffer, offset=0):
        self._buffer = buffer;
        self._offset = offset;

    def load_32bit_number(self):
        '''
        :return: number
        '''
        result = struct.unpack_from('<I', self._buffer, self._offset)[0]
        self._offset += 4
        return result

    def load_16bit_number(self):
        '''
        :return: number
        '''
        result = struct.unpack_from('<H', self._buffer, self._offset)[0]
        self._offset += 2
        return result

    def load_string(self):
        str_length = self.load_16bit_number()
        offset = self._offset
        data = self._buffer
        if (str_length > 32767):
            str_length = str_length - 32768
            result = data[offset : offset + str_length].decode('latin-1')
            self._offset += str_length + str_length % 2
        else:
            result = data[offset : offset + str_length * 2].decode('utf_16_le')
            self._offset += str_length * 2
        return result

    def load_string_list(self):
        length = self.load_32bit_number()
        return [self.load_string() for i in _range(length)]

    def load_string_list_map(self):
        result = {}
        length = self.load_32bit_number()
        for i in _range(length):
            key = self.load_string()
            values = self.load_string_list()
            result[key] = values
        return result

    def load_32bit_number_list(self):
        result_length = self.load_32bit_number()
        result = []
        while len(result) < result_length:
            tag = self.load_16bit_number()
            if (tag >> 15) == 1: # zebra
                length = min(result_length - len(result), 15)
                for i in _range(length):
                    if (tag >> i) & 0x1:
                        result.append(self.load_32bit_number())
                    else:
                        result.append(0);
            elif (tag >> 14) == 1: # non-zero
                length = tag - 0x4000 + 1
                for i in _range(length):
                    result.append(self.load_32bit_number())
            else: #zero
                length = tag + 1
                for i in _range(length):
                    result.append(0)
        return result

class BinaryOutput(object):
    def __init__(self):
        self._output = []
        self.outputBytes = 0;

    def dump_32bit_number(self, num):
        self._output.append(struct.pack('<I', num % (2 ** 32)))
        self.outputBytes += 4

    def dump_16bit_number(self, num):
        self._output.append(struct.pack('<H', num % (2 ** 16)))
        self.outputBytes += 2

    @staticmethod
    def convert_16bit_number(num):
        return struct.pack('<H', num % (2 ** 16))

    def dump_string(self, string):
        if type(string) is list:
            string = struct.pack('<%dH' % len(string), *string).decode("utf_16_le")
        converted_result = BinaryOutput.convert_string(string)
        self.outputBytes += len(converted_result)
        self._output.append(converted_result)

    @staticmethod
    def convert_string(string):
        result = []
        if len(string) > 32768:
            string = string[0:32768]
        length = len(string)
        byte_str = string.encode('utf_16_le')
        compress = True
        char_codes = []
        for i in _range(1, len(byte_str), 2):
            if byte_str[i] != b'\x00' and byte_str[i] != 0:
                compress = False
                break;
        if compress:
            result.append(BinaryOutput.convert_16bit_number(length + 32768))
            result.append(string.encode('latin-1'))
            if length % 2 == 1:
                result.append(b'\x00')
        else:
            result.append(BinaryOutput.convert_16bit_number(length))
            result.append(byte_str)
        return b"".join(result)

    def dump_raw_string(self, string):
        self._output.append(string)
        self.outputBytes += len(string)

    def dump_string_list(self, str_list):
        self.dump_32bit_number(len(str_list))
        for string in str_list:
            self.dump_string(string)

    def dump_string_list_map(self, str_map):
        self.dump_32bit_number(len(str_map))
        for key, value in str_map.items():
            self.dump_string(key)
            self.dump_string_list(value)

    def dump_32bit_number_list(self, array):
        self.dump_32bit_number(len(array))
        index = 0;
        input_length = len(array)
        while index < input_length:
            if array[index] == 0:
                length = self._count_zero(array, index)
                self._zero_block(length)
                index += length
            elif (self._should_zebra_code(array, index)):
                self._create_zebra_code(array, index)
                index = min(len(array), index + 15)
            else:
                length = self._search_double_zero(array, index)
                self._non_zero_block(array, index, length)
                if length == 0:
                    raise ValueError()
                index += length

    def _count_zero(self, array, offset):
        for i in _range(offset, len(array)):
            if array[i] != 0:
                return i - offset
        return len(array) - offset

    def _zero_block(self, length):
        while length > 0:
            if length > 16384:
                self.dump_16bit_number(16384 - 1)
                length -= 16384
            else:
                self.dump_16bit_number(length - 1)
                length = 0

    def _should_zebra_code(self, array, offset):
        if len(array) - offset < 16:
            return True
        change = 0
        is_last_zero = False
        for i in _range(offset, offset + 15):
            if array[i] == 0:
                if not is_last_zero:
                    is_last_zero = True
                    change += 1
            else:
                if is_last_zero:
                    is_last_zero = False
                    change += 1
        return change > 2

    def _search_double_zero(self, array, offset):
        is_last_zero = False;
        for i in _range(offset, len(array)):
            if array[i] == 0:
                if is_last_zero:
                    return i - offset - 1
                is_last_zero = True
            else:
                is_last_zero = False
        return len(array) - offset

    def _non_zero_block (self, array, offset, length):
        while length > 0:
            if length > 16384:
                block_length = 16384
                length -= 16384
            else:
                block_length = length
                length = 0
            self.dump_16bit_number((block_length - 1) + 0x4000)
            for i in _range(offset, offset + block_length):
                self.dump_32bit_number(array[i])
            offset += block_length

    def _create_zebra_code(self, array, offset):
        last = min(offset + 15, len(array))
        code = 0x8000
        temp_output = self._output
        index = len(self._output)
        for i in _range(offset, last):
            if array[i] != 0:
                self.dump_32bit_number(array[i]);
                code = code + (0x1 << (i - offset));
        self._output.insert(index, struct.pack('<H', code % (2 ** 16)))
        self.outputBytes += 2

    def result(self):
        return b''.join(self._output);



import sys
import argparse
import snowballstemmer
from . import oktavia
from . import htmlparser
from . import csvparser
from . import textparser
from . import stemmer

def usage():
    print(''.join([
        "usage: oktavia-mkindex-cli [options]",
        "",
        "Common Options:",
        " -i, --input [input folder/file ] : Target files to search. .html, .csv, .txt are available.",
        " -o, --output [outputfolder]      : Directory that will store output files.",
        "                                  : This is a relative path from root.",
        "                                  : Default value is 'search'. ",
        " -t, --type [type]                : Export type. 'index', 'base64', 'cmd', 'js(default)',",
        "                                  : 'commonjs' 'web', are available.",
        "                                  : 'index' is a just index file. 'cmd' is a base64 code with search program.",
        "                                  : Others are base64 source code style output.",
        " -m, --mode [mode]                : Mode type. 'html', 'csv', 'text' are available.",
        " -c, --cache-density [percent]    : Cache data density. It effects file size and search speed.",
        "                                  : 100% become four times of base index file size. Default value is 5%.",
        "                                  : Valid value is 0.1% - 100%.",
        " -n, --name [function]            : A variable name for 'js' output or property name",
        "                                  : for 'js' and 'commonjs'. Default value is 'searchIndex'.",
        " -q, --quiet                      : Hide detail information.",
        " -h, --help                       : Display this message.",
        "",
        "HTML Mode Options:",
        " -r, --root  [document root]      : Document root folder. Default is current. ",
        "                                  : Indexer creates result file path from this folder.",
        " -p, --prefix [directory prefix]  : Directory prefix for a document root from a server root.",
        "                                  : If your domain is example.com and 'manual' is passed,",
        "                                  : document root become http://example.com/manual/.",
        "                                  : It effects search result URL. Default value is '/'.",
        " -u, --unit [search unit]         : 'file', 'h1'-'h6'. Default value is 'file'.",
        " -f, --filter [target tag]        : Only contents inside this tag is indexed.",
        "                                  : Default value is \"article,#content,#main,div.body\".",
        " -s, --stemmer [algorithm]        : Select stemming algorithm.",
        " -w, --word-splitter [splitter]   : Use optional word splitter.",
        "                                  : 'ts' (TinySegmenter for Japanese) is available",
        "",
        "Text Mode Options:",
        " -r, --root  [document root]      : Document root folder. Default is current. ",
        "                                  : Indexer creates result file path from this folder.",
        " -s, --stemmer [algorithm]        : Select stemming algorithm.",
        " -w, --word-splitter [splitter]   : Use optional word splitter.",
        "                                  : 'ts' (TinySegmenter for Japanese) is available",
        " -u, --unit [search unit]         : file, block, line. Default value is 'file'.",
        "",
        "Supported Stemmer Algorithms:",
        "  danish, dutch, english, finnish, french german, hungarian italian",
        "  norwegian, porter, portuguese, romanian, russian, spanish, swedish, turkish"
    ]))

def main():
    print("Search Engine Oktavia - Index Generator\n")

    parser = argparse.ArgumentParser(description='Search Engine Oktavia - Index Generator')
    inputs = [] : string[]
    root = process.cwd()
    prefix = '/'
    output : Nullable.<string> = null
    showhelp = false
    notrun = false
    unit = 'file'
    type = 'js'
    mode = ''
    verbose = true
    filter = [] : string[]
    algorithm : Nullable.<string> = null
    //wordsplitter : Nullable.<string> = null
    cacheDensity : number = 5.0
    name = null : Nullable.<string>
    validModes = ['html', 'csv', 'text']
    validUnitsForHTML = ['file', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6']
    validUnitsForText = ['file', 'block', 'line']
    validStemmers = [
        'danish', 'dutch', 'english', 'finnish', 'french', 'german', 'hungarian',
        'italian', 'norwegian', 'porter', 'portuguese', 'romanian', 'russian',
        'spanish', 'swedish', 'turkish'
    ]
    validTypes = ['index', 'base64', 'cmd', 'js', 'commonjs', 'web']
    //validWordSplitters = ['ts']

    optstring = "n:(name)q(quiet)m:(mode)i:(input)r:(root)p:(prefix)o:(output)h(help)u:(unit)f:(filter)s:(stemmer)w:(word-splitter)t:(type)c:(cache-density)"
    parser = BasicParser(optstring, args)
    opt = parser.getopt()
    while (opt)
    {
        switch (opt.option)
        {
        case "m":
            if (validModes.indexOf(opt.optarg) == -1)
            {
                console.error("Option m/mode should be 'html', 'csv', 'text'.")
                notrun = true
            }
            mode = opt.optarg
            break
        case "i":
            inputs.push(opt.optarg)
            break
        case "r":
            root = node.path.resolve(opt.optarg)
            break
        case "p":
            prefix = opt.optarg
            break
        case "n":
            name = opt.optarg
            break
        case "o":
            output = opt.optarg
            if (output.slice(0, 1) == '/')
            {
                output = output.slice(1)
            }
            break
        case "h":
            showhelp = true
            break
        case "q":
            verbose = false
            break
        case "u":
            unit = opt.optarg
            break
        case "f":
            items = opt.optarg.split(',')
            for (i in items)
            {
                filter.push(items[i])
            }
            break
        case "t":
            if (validTypes.indexOf(opt.optarg) == -1)
            {
                console.error('Option -t/--type is invalid.')
                notrun = true
            }
            else
            {
                type = opt.optarg
            }
            break
        case "s":
            if (validStemmers.indexOf(opt.optarg) == -1)
            {
                console.error('Option -s/--stemmer is invalid.')
                notrun = true
            }
            else
            {
                algorithm = opt.optarg
            }
            break
        /*case "w":

            break*/
        case "c":
            match = /(\d+\.?\d*)/.exec(opt.optarg)
            if (match)
            {
                tmpValue = match[1] as number
                if (0.1 <= tmpValue && tmpValue <= 100)
                {
                    cacheDensity = tmpValue
                }
                else
                {
                    console.error('Option -c/--cache-density should be in 0.1 - 100.')
                    notrun = true
                }
            }
            else
            {
                console.error('Option -c/--cache-density is invalid.')
                notrun = true
            }
            break
        case "?":
            notrun = true
            break
        }
        opt = parser.getopt()
    }
    inputTextFiles = [] : string[]
    inputHTMLFiles = [] : string[]
    inputCSVFiles = [] : string[]
    if (filter.length == 0)
    {
        filter = ['article', '#content', '#main', 'div.body']
    }
    for (i in inputs)
    {
        input = inputs[i]
        if (!node.fs.existsSync(input))
        {
            console.error("Following input folder/file doesn't exist: " + input)
            notrun = true
        }
        else
        {
            stat = node.fs.statSync(input)
            if (stat.isFile())
            {
                _Main._checkFileType(node.path.resolve(input), inputTextFiles, inputHTMLFiles, inputCSVFiles)
            }
            else if (stat.isDirectory())
            {
                _Main._checkDirectory(input, inputTextFiles, inputHTMLFiles, inputCSVFiles)
            }
            else
            {
                console.error("Following input is not folder or file: " + input)
                notrun = true
            }
        }
    }
    if (inputTextFiles.length == 0 && inputHTMLFiles.length == 0 && inputCSVFiles.length == 0 || !mode)
    {
        showhelp = true
    }
    if (showhelp)
    {
        _Main.usage()
    }
    else if (!notrun)
    {
        stemmer : Nullable.<Stemmer> = null
        if (algorithm)
        {
            stemmer = _Main._createStemmer(algorithm)
        }
        dump = null : Nullable.<string>
        switch (mode)
        {
        case 'html':
            unitIndex = validUnitsForHTML.indexOf(unit)
            if (unitIndex == -1)
            {
                console.error('Option -u/--unit should be file, h1, h2, h3, h4, h5, h6. But ' + unit)
            }
            else
            {
                htmlParser = HTMLParser(unitIndex, root, prefix, filter, stemmer)
                for (i = 0 i < inputHTMLFiles.length i++)
                {
                    htmlParser.parse(inputHTMLFiles[i])
                }
                console.log('generating index...')
                if (verbose)
                {
                    console.log('')
                }
                dump = htmlParser.dump(cacheDensity, verbose)
            }
            break
        case 'csv':
            csvParser = CSVParser(root, stemmer)
            for (i in inputCSVFiles)
            {
                csvParser.parse(inputCSVFiles[i])
            }
            break
        case 'text':
            if (validUnitsForText.indexOf(unit) == -1)
            {
                console.error('Option u/unit should be file, block, line. But ' + unit)
            }
            else
            {
                textParser = TextParser(unit, root, stemmer)
                for (i in inputTextFiles)
                {
                    textParser.parse(inputTextFiles[i])
                }
                dump = textParser.dump(cacheDensity, verbose)
            }
            break
        }
        if (dump)
        {
            indexFilePath = ""
            switch (type)
            {
            case 'index':
                if (output == null)
                {
                    output = 'search'
                }
                indexFilePath = node.path.resolve(root, output, 'searchindex.okt')
                dirPath = node.path.dirname(indexFilePath)
                _Main._mkdirP(dirPath)
                console.log("writing index in binary...", indexFilePath)
                node.fs.writeFileSync(indexFilePath, dump, "utf16le")
                break
            case 'base64':
                if (output == null)
                {
                    output = 'search'
                }
                indexFilePath = node.path.resolve(root, output, 'searchindex.okt.b64')
                dirPath = node.path.dirname(indexFilePath)
                _Main._mkdirP(dirPath)
                console.log("writing index in base64...", indexFilePath)
                node.fs.writeFileSync(indexFilePath, Base64.btoa(dump), "utf8")
                break
            case 'cmd':
                if (output == null)
                {
                    output = 'index-search-cli'
                }
                srcLines = node.fs.readFileSync(node.path.join(node.__dirname, 'oktavia-cli-runtime.js'), 'utf8').split('\n')
                latinString = Base64.to8bitString(dump)
                base64String = Base64.btoa(latinString)
                indexContent = 'searchIndex = "' + base64String + '"'
                srcLines.splice(1, 0, indexContent)
                console.log("writing index for cli...", output)
                node.fs.writeFileSync(output, srcLines.join('\n'), 'utf8')
                node.fs.chmodSync(output, '0755')
                break
            case 'web':
                if (output == null)
                {
                    output = 'searchindex.js'
                }
                srcLines = node.fs.readFileSync(node.path.join(node.__dirname, "web", 'oktavia-web-runtime.js'), 'utf8').split('\n')
                latinString = Base64.to8bitString(dump)
                base64String = Base64.btoa(latinString)
                indexContent = 'searchIndex = "' + base64String + '"'
                srcLines.splice(0, 0, indexContent)
                console.log("writing index for web...", output)
                node.fs.writeFileSync(output, srcLines.join('\n'), 'utf8')
                break
            case 'js':
                if (output == null)
                {
                    output = 'search'
                }
                indexFilePath = node.path.resolve(root, output, 'searchindex.js')
                dirPath = node.path.dirname(indexFilePath)
                _Main._mkdirP(dirPath)
                if (name == null)
                {
                    name = 'searchIndex'
                }
                contents = [
                    '// Oktavia Search Index',
                    '' + name + ' = "' + Base64.btoa(dump) + '"', ''
                ]
                console.log("writing index in js...", indexFilePath)
                node.fs.writeFileSync(indexFilePath, contents.join('\n'), "utf8")
                break
            case 'commonjs':
                if (output == null)
                {
                    output = 'search'
                }
                indexFilePath = node.path.resolve(root, output, 'searchindex.js')
                dirPath = node.path.dirname(indexFilePath)
                _Main._mkdirP(dirPath)
                if (name == null)
                {
                    name = 'searchIndex'
                }
                contents = [
                    '// Oktavia Search Index',
                    'exports.' + name + ' = "' + Base64.btoa(dump) + '"', ''
                ]
                console.log("writing index in js...", indexFilePath)
                node.fs.writeFileSync(indexFilePath, contents.join('\n'), "utf8")
                break
            }
        }
    }
}

def _checkFileType (path : string, texts : string[], HTMLs : string[], CSVs : string[]) : void
{
    match = path.match(/(.*)\.(.*)/)
    if (match && match[1])
    {
        switch (match[2].toLowerCase())
        {
        case 'html':
        case 'htm':
            HTMLs.push(path)
            break
        case 'csv':
            CSVs.push(path)
            break
        default:
            texts.push(path)
        }
    }
}

def checkDirectory (path : string, texts : string[], HTMLs : string[], CSVs : string[]) : void
{
    files = node.fs.readdirSync(path)
    for (j in files)
    {
        filepath = node.path.resolve(path, files[j])
        stat = node.fs.statSync(filepath)
        if (stat.isFile())
        {
            _Main._checkFileType(filepath, texts, HTMLs, CSVs)
        }
        else if (stat.isDirectory())
        {
            _Main._checkDirectory(filepath, texts, HTMLs, CSVs)
        }
    }
}

def mkdirP (path : string) : void
{
    if (node.fs.existsSync(path))
    {
        return
    }
    _Main._mkdirP(node.path.dirname(path))
    node.fs.mkdirSync(path)
}

def createStemmer(algorithm):
    stemmer : Stemmer
    switch (algorithm.toLowerCase())
    {
    case "danish":
        stemmer = DanishStemmer()
        break
    case "dutch":
        stemmer = DutchStemmer()
        break
    case "english":
        stemmer = EnglishStemmer()
        break
    case "finnish":
        stemmer = FinnishStemmer()
        break
    case "french":
        stemmer = FrenchStemmer()
        break
    case "german":
        stemmer = GermanStemmer()
        break
    case "hungarian":
        stemmer = HungarianStemmer()
        break
    case "italian":
        stemmer = ItalianStemmer()
        break
    case "norwegian":
        stemmer = NorwegianStemmer()
        break
    case "porter":
        stemmer = PorterStemmer()
        break
    case "portuguese":
        stemmer = PortugueseStemmer()
        break
    case "romanian":
        stemmer = RomanianStemmer()
        break
    case "russian":
        stemmer = RussianStemmer()
        break
    case "spanish":
        stemmer = SpanishStemmer()
        break
    case "swedish":
        stemmer = SwedishStemmer()
        break
    case "turkish":
        stemmer = TurkishStemmer()
        break
    default:
        stemmer = EnglishStemmer()
        break
    }
    return stemmer



import re
import math
import struct
import snowballstemmer
from . import fmindex
from . import binaryio
from . import query
from . import searchresult
from . import metadata

if hasattr(__builtins__, 'unichr'):
    tostr = unichr
else:
    tostr = chr

wordsplitter = re.compile(r'\s')
wordsplitter2 = re.compile(r'\W')

def to_utf16(string):
    binary = string.encode('utf-16le')
    return struct.unpack("%dH" % len(string), binary)

def from_utf16(codes):
    return struct.pack("%dH" % len(codes), *codes).decode('utf-16le')

class Oktavia(object):
    # sentinels
    eof = 0
    eob = 1
    unknown = 2

    # Enum
    USE_STEMMING = True

    def __init__(self):
        self._fmindex = fmindex.FMIndex(rawmode=True)
        self._metadatas = {}
        self._metadataLabels = []
        self._stemmer = None
        self._stemmingResult = {}
        self._build = False
        self._isLastEob = False
        self._utf162compressCode = {Oktavia.eof: 0, Oktavia.eob: 1, Oktavia.unknown: 2}
        self._compressCode2utf16 = [Oktavia.eof, Oktavia.eob, Oktavia.unknown]

    def set_stemmer(self, stemmer):
        self._stemmer = stemmer

    def get_primary_metadata(self):
        return self._metadatas[self._metadataLabels[0]]

    def add_section(self, key):
        if key in self._metadataLabels:
            raise ValueError('Metadata name ' + key + ' is already exists')
        self._metadataLabels.append(key)
        section = metadata.Section(self, key)
        self._metadatas[key] = section
        return section

    def get_section(self, key):
        if key not in self._metadataLabels:
            raise ValueError('Metadata name ' + key + " does't exists")
        return self._metadatas[key]

    def add_splitter(self, key):
        if key in self._metadataLabels:
            raise ValueError('Metadata name ' + key + ' is already exists')
        self._metadataLabels.append(key)
        splitter = metadata.Splitter(self, key)
        self._metadatas[key] = splitter
        return splitter

    def get_splitter(self, key):
        if key not in self._metadataLabels:
            raise ValueError('Metadata name ' + key + " does't exists")
        return self._metadatas[key]

    def add_table(self, key, headers):
        if key in self._metadataLabels:
            raise ValueError('Metadata name ' + key + ' is already exists')
        self._metadataLabels.append(key)
        table = metadata.Table(self, key, headers)
        self._metadatas[key] = table
        return table

    def get_table(self, key):
        if key not in self._metadataLabels:
            raise ValueError('Metadata name ' + key + " does't exists")
        return self._metadatas[key]

    def add_block(self, key):
        if key in self._metadataLabels:
            raise ValueError('Metadata name ' + key + ' is already exists')
        self._metadataLabels.append(key)
        block = metadata.Block(self, key)
        self._metadatas[key] = block
        return block

    def get_block(self, key):
        if key not in self._metadataLabels:
            raise ValueError('Metadata name ' + key + " does't exists")
        return self._metadatas[key]

    def add_end_of_block(self):
        self._fmindex.append([Oktavia.eob])
        self._isLastEob = True

    def add_word(self, word, stemming=None):
        word.encode
        string = []
        utf16word = to_utf16(word)
        for charCode in utf16word:
            if charCode not in self._utf162compressCode:
                convertedChar = len(self._compressCode2utf16)
                self._utf162compressCode[charCode] = convertedChar
                self._compressCode2utf16.append(charCode)
            else:
                convertedChar = self._utf162compressCode[charCode]
            string.append(convertedChar)
        self._fmindex.append(string)
        self._isLastEob = utf16word[-1] < 2

        if stemming is None:
            return

        wordList = wordsplitter.split(word)
        for originalWord in wordList:
            smallWord = originalWord[0:1].lower() + originalWord[1:]
            registerWord = None
            if stemming and self._stemmer:
                baseWord = self._stemmer.stemWord(originalWord.lower())
                if baseWord not in originalWord:
                    registerWord = baseWord
            elif originalWord != smallWord:
                registerWord = smallWord
            if registerWord:
                compressedCodeWord = self._convertToCompressionCode(originalWord)
                if registerWord not in self._stemmingResult:
                    stemmedList = [compressedCodeWord]
                    self._stemmingResult[registerWord] = stemmedList
                else:
                    stemmedList = self._stemmingResult[registerWord]
                    if compressedCodeWord not in stemmedList:
                        stemmedList.append(compressedCodeWord)

    def _convertToCompressionCode(self, keyword):
        resultChars = []
        if type(keyword) is not list:
            keyword = to_utf16(keyword)
        for originalChar in keyword:
            if originalChar not in self._utf162compressCode:
                resultChars.append("\x02")
            else:
                char = self._utf162compressCode[originalChar]
                resultChars.append(char)
        return resultChars

    def raw_search(self, keyword, stemming=False):
        if not self._build:
            raise RuntimeError("Oktavia.build() is not called yet")
        if stemming:
            result = []
            if self._stemmer:
                baseWord = self._stemmer.stemWord(keyword.lower())
                if baseWord in self._stemmingResult:
                    stemmedList = self._stemmingResult[baseWord]
                    for word in stemmedList:
                        result += self._fmindex.search(word)
        else:
            result = self._fmindex.search(self._convertToCompressionCode(keyword))
        return result

    def search(self, queries):
        if not self._build:
            raise RuntimeError("Oktavia.build() is not called yet")
        summary = SearchSummary(self)
        for query in queries:
            summary.add_query(self._searchQuery(query))
        summary.merge_result()
        return summary

    def _searchQuery(self, query):
        result = SingleResult(query.word, query.OR, query.NOT)
        if query.raw:
            positions = self.raw_search(query.word, False)
        else:
            positions = self.raw_search(query.word, False).concat(self.rawSearch(query.word, True))
        self.getPrimaryMetadata().grouping(result, positions, query.word, not query.raw)
        return result

    def build(self, cacheDensity=5):
        if self._build:
            raise RuntimeError("Oktavia.build() is already called")
        for key in self._metadatas:
            self._metadatas[key]._build()
        maxCharCode = len(self._compressCode2utf16)
        cacheRange = round(max(1, (100 / min(100, max(0.01, cacheDensity)))))
        self._fmindex.build(cacheRange, maxCharCode)
        self._build = True

    def dump(self, verbose=False):
        if not self._build:
            raise RuntimeError("Oktavia.build() is not called yet")
        output = binaryio.BinaryOutput()
        headerSource = u"oktavia-02"
        output.dump_raw_string(binaryio.BinaryOutput.convert_string(headerSource)[2:])
        if verbose:
            print('Source text size: %d bytes' % (self._fmindex.size() * 2))
        self._fmindex.dump(output)
        output.dump_string(self._compressCode2utf16[3:])
        if verbose:
            print('Char Code Map: %d bytes' % (len(self._compressCode2utf16) * 2 - 2))
        size = output.outputBytes
        output.dump_string_list_map(self._stemmingResult)
        if verbose:
            print('Stemmed Word Table: %d bytes' % (output.outputBytes - size) * 2)

        output.dump_16bit_number(len(self._metadataLabels))
        for name in self._metadataLabels:
            size = output.outputBytes
            self._metadatas[name]._dump(output)
            if verbose:
                print('Meta Data %s: %d bytes' % (name, output.outputBytes - size) * 2)
        return output.result()

    def load(self, data):
        headerSource = u"oktavia-02"
        header = binaryio.BinaryOutput.convert_string(headerSource)[2:]
        if data[0:10] != header:
            raise ValueError('Invalid data file')
        input = binaryio.BinaryInput(data, 10)
        self._metadatas = {}
        self._metadataLabels = []

        self._fmindex.load(input)
        charCodes = input.load_string()
        self._utf162compressCode = {Oktavia.eof: 0, Oktavia.eob: 1, Oktavia.unknown: 2}
        self._compressCode2utf16 = [Oktavia.eof, Oktavia.eob, Oktavia.unknown]

        for i, charCode in enumerate(to_utf16(charCodes)):
            self._compressCode2utf16.append(charCode)
            self._utf162compressCode[charCode] = i + 3

        self._stemmingResult = input.load_string_list_map()

        metadataCount = input.load_16bit_number()
        for i in range(metadataCount):
            typecode = input.load_16bit_number()
            if typecode == metadata.Section.TypeID:
                metadata.Section._load(self, input)
            elif typecode == metadata.Splitter.TypeID:
                metadata.Splitter._load(self, input)
            elif typecode == metadata.Table.TypeID:
                metadata.Table._load(self, input)
            elif typecode == metadata.Block.TypeID:
                metadata.Block._load(self, input)
            else:
                raise TypeError("Metadata TypeError: %d" % type)
        self._build = True

    def content_size(self):
        return self._fmindex.content_size()

    def word_position_type(self, position):
        result = 0
        if position == 0:
            result = 4
        else:
            ahead = self._fmindex.get_substring(position - 1, 1)
            if wordsplitter.match(ahead):
                result = 2
            elif wordsplitter2.match(ahead):
                result = 1
            elif Oktavia.eob == ahead:
                result = 3
        return result

    def _get_substring(self, position, length):
        codes = self._fmindex.get_substring(position, length)
        result = [] 
        for code in codes:
            if code > 2:
                result.append(self._compressCode2utf16[code])
        return from_utf16(result)

    def _get_substring_with_EOB(self, position, length):
        result = self._fmindex.get_substring(position, length)
        str = []
        for code in result:
            if code > 2:
                str.append(self._compressCode2utf16[code])
            elif code == 1:
                str.append(1)
        return from_utf16(str)



class Proposal(object):
    def __init__(self, omit, expect):
        self.omit = omit
        self.expect = expect


class Position(object):
    def __init__(self, word, position, stemmed):
        self.word = word
        self.position = position
        self.stemmed = stemmed


class SearchUnit(object):
    def __init__ (self, id):
        self.positions = {}
        self.id = id
        self._size = 0
        self.score = 0
        self.startPosition = -1

    def add_position(self, word, position, stemmed):
        if position not in self.positions: 
            self._size += 1
            self.positions[position] = Position(word, position, stemmed)
        else:
            positionObj = self.positions[position]
            if len(positionObj.word) < len(word):
                positionObj.word = word
            positionObj.stemmed = positionObj.stemmed and stemmed

    def get(self, position):
        return self.positions[position]

    def __len__(self):
        return self._size

    def merge(self, rhs):
        for pos in rhs.positions:
            self.addPosition(pos.word, pos.position, pos.stemmed)

    def get_positions(self):
        result = []
        for pos in self.positions:
            result.append(pos)
        result.sort(key=lambda a: -a.position)
        return result


class SingleResult(object):
    def __init__(self, searchWord='', OR=False, NOT=False):
        self.units = []
        self.unitIds = []
        self.OR = OR
        self.NOT = NOT
        self.searchWord = searchWord

    def get_search_unit(self, unitId):
        if unitId not in self.unitIds:
            result = SearchUnit(unitId)
            self.units.append(result)
            self.unitIds.append(unitId)
        else:
            result = self.units[self.unitIds.index(unitId)]
        return result

    def merge(self, rhs):
        result = SingleResult()
        if rhs.OR:
            self._orMerge(result, rhs)
        elif rhs.NOT:
            self._notMerge(result, rhs)
        else:
            self._andMerge(result, rhs)
        return result

    def __len__(self):
        return len(self.units)

    def _andMerge(self, result, rhs):
        for i, id in enumerate(self.unitIds):
            if id in rhs.unitIds:
                lhsSection = self.units[i]
                result.unitIds.append(id)
                result.units.append(lhsSection)

    def _orMerge(self, result, rhs):
        result.unitIds = self.unitIds[:]
        result.units = self.units[:]

        for i, id in enumerate(rhs.unitIds):
            rhsSection = rhs.units[i]
            if id in result.unitIds:
                lhsSection = result.units[result.unitIds.index(id)]
                lhsSection.merge(rhsSection)
            else:
                result.unitIds.append(id)
                result.units.append(rhsSection)

    def _notMerge(self, result, rhs):
        for i, id in enumerate(self.unitIds):
            if id not in rhs.unitIds:
                lhsSection = self.units[i]
                result.unitIds.append(id)
                result.units.append(lhsSection)


class SearchSummary(object):
    def __init__ (self, oktavia=None):
        self.sourceResults = []
        self.result = None
        self.oktavia = oktavia

    def add_query(self, result):
        self.sourceResults.append(result)

    def merge_result(self, results=None):
        if results is None:
            self.result = self.merge_result(self.sourceResults)
        else:
            rhs = results[0]
            for result in results:
                rhs = rhs.merge(result)
            return rhs

    def get_proposals(self):
        proposals = []
        for i in range(len(self.sourceResults)):
            tmpSource = []
            for j in range(len(self.sourceResults)):
                if i != j:
                    tmpSource.append(self.sourceResults[j])
            result = self.merge_result(tmpSource)
            proposals.append(Proposal(i, len(result)))
        
        proposals.sort(key=lambda a: -a.expect)
        return proposals

    def get_sorted_result(self):
        result = self.result.units[:]
        result.sort(key=lambda a: -a.score)
        return result

    def __len__(self):
        return len(self.result)

    def add(self, result):
        self.sourceResults.append(result)



'''
Original source code:
*  G. Nong, S. Zhang and W. H. Chan, Two Efficient Algorithms for Linear Time Suffix Array Construction, IEEE Transactions on Computers, To Appear
* http:#www.cs.sysu.edu.cn/nong/index.files/Two%20Efficient%20Algorithms%20for%20Linear%20Suffix%20Array%20Construction.pdf
'''

from . bitvector import BitVector

_range = getattr(__builtins__, 'xrange', range)

class OArray(object):
    def __init__(self, array, offset = 0):
        self.array = array
        self.offset = offset

    def get(self, index):
        return self.array[index + self.offset]

    def set(self, index, value):
        self.array[index + self.offset] = value

    def isS(self, index):
        return self.array[index + self.offset] < self.array[index + self.offset + 1]

    def compare(self, index1, index2):
        return self.array[index1 + self.offset] == self.array[index2 + self.offset]

class SAIS(object):
    @staticmethod
    def _isLMS(t, i):
        return i > 0 and t.get(i) and not t.get(i - 1)

    @staticmethod
    def _getBuckets(s, bkt, n, K, end):
        '''find the start or end of each bucket'''
        sum = 0
        for i in _range(K + 1):
            bkt[i] = 0 # clear all buckets
        for i in _range(n):
            bkt[s.get(i)] += 1 # compute the size of each bucket
        for i in _range(K + 1):
            sum += bkt[i]
            bkt[i] = sum if end else sum - bkt[i]

    @staticmethod
    def _induceSAl(t, SA, s, bkt, n, K, end):
        # compute SAl
        SAIS._getBuckets(s, bkt, n, K, end) # find starts of buckets
        for i in _range(n):
            j = SA[i] - 1
            if j >= 0 and not t.get(j):
                index = s.get(j)
                SA[bkt[index]] = j
                bkt[index] += 1

    @staticmethod
    def _induceSAs(t, SA, s, bkt, n, K, end):
        # compute SAs
        SAIS._getBuckets(s, bkt, n, K, end) # find ends of buckets
        for i in _range(n - 1, -1, -1):
            j = SA[i] - 1
            if j >= 0 and t.get(j):
                index = s.get(j)
                bkt[index] -= 1
                SA[bkt[index]] = j

    @staticmethod
    def make(source, rawmode=False):
        '''
        find the suffix array SA of s[0..n-1] in 1..K^n
        require s[n-1]=0 (the sentinel!), n>=2
        use a working space (excluding s and SA) of at most 2.25n+O(1) for a constant alphabet
        '''
        if rawmode:
            charCodes = source
        else:
            charCodes = [ord(c) for c in source]
        
        SA = [0] * len(source)
        s = OArray(charCodes)
        SAIS._make(s, SA, len(source), max(charCodes))
        return SA

    @staticmethod
    def _make(s, SA, n, K):
        # Classify the type of each character
        t = BitVector()
        t.set(n - 2, False)
        t.set(n - 1, True) # the sentinel must be in s1, important!!!
        for i in _range(n - 3, -1, -1):
            t.set(i, s.isS(i) or (s.compare(i, i + 1) and t.get(i + 1)))
        # stage 1: reduce the problem by at least 1/2
        # sort all the S-substrings
        bkt = [0] * (K + 1)
        SAIS._getBuckets(s, bkt, n, K, True) # find ends of buckets
        for i in _range(n):
            SA[i] = -1
        for i in _range(1, n):
            if SAIS._isLMS(t, i):
                index = s.get(i)
                bkt[index] -= 1
                SA[bkt[index]] = i
        SAIS._induceSAl(t, SA, s, bkt, n, K, False)
        SAIS._induceSAs(t, SA, s, bkt, n, K, True)
        # compact all the sorted substrings into the first n1 items of SA
        # 2*n1 must be not larger than n (proveable)
        n1 = 0
        for i in _range(n):
            if SAIS._isLMS(t, SA[i]):
                SA[n1] = SA[i]
                n1 += 1

        # find the lexicographic names of all substrings
        for i in _range(n1, n):
            SA[i]=-1 # init the name array buffer
        
        name = 0
        prev = -1
        for i in _range(n1):
            pos = SA[i]
            diff = False
            for d in _range(n):
                if prev == -1 or not s.compare(pos + d, prev + d) or t.get(pos + d) != t.get(prev + d):
                    diff = True
                    break
                elif d > 0 and (SAIS._isLMS(t, pos+d) or SAIS._isLMS(t, prev + d)):
                    break
            if diff:
                name += 1
                prev = pos
            pos = pos // 2 if (pos % 2 == 0) else (pos - 1) // 2
            SA[n1 + pos] = name - 1
       
        j = n - 1
        for i in _range(n - 1, n1 - 1, -1):
            if SA[i] >= 0:
                SA[j] = SA[i]
                j -= 1
        # stage 2: solve the reduced problem
        # recurse if names are not yet unique
        SA1 = SA
        s1 = OArray(SA, n - n1)
        if name < n1:
            SAIS._make(s1, SA1, n1, name - 1)
        else:
            # generate the suffix array of s1 directly
            for i in _range(n1):
                SA1[s1.get(i)] = i
        # stage 3: induce the result for the original problem
        bkt = [0] * (K + 1)
        # put all left-most S characters into their buckets
        SAIS._getBuckets(s, bkt, n, K, True) # find ends of buckets
        j = 0
        for i in _range(n):
            if SAIS._isLMS(t, i):
                s1.set(j, i) # get p1
                j += 1
        for i in _range(n1):
            SA1[i] = s1.get(SA1[i]) # get index in s
        for i in _range(n1, n):
            SA[i] = -1 # init SA[n1..n-1]
        for i in _range(n1 - 1, -1, -1):
            j = SA[i]
            SA[i] = -1
            index = s.get(j)
            bkt[index] -= 1
            SA[bkt[index]] = j
        SAIS._induceSAl(t, SA, s, bkt, n, K, False)
        SAIS._induceSAs(t, SA, s, bkt, n, K, True)




'''
This is a JSX version of shellinford library:
https://code.google.com/p/shellinford/

License: http://shibu.mit-license.org/
'''

import copy
import math
import struct

from . import binaryio
from . import bitvector

_range = getattr(__builtins__, 'xrange', range)

class WaveletMatrix(object):

    def __init__(self):
        self._range = {}
        self._bv = []
        self._seps = []
        self._maxcharcode = 65535
        self._bitsize = 16
        self._usedChars = set()
        self.clear()

    def bitsize(self):
        return self._bitsize

    def set_max_char_code(self, char_code):
        self._maxcharcode = char_code
        self._bitsize = int(math.ceil(math.log(self._maxcharcode) / math.log(2)))

    def max_char_code(self):
        return self._maxcharcode

    def clear(self):
        del self._bv[:]
        del self._seps[:]
        self._size = 0

    def usedChars(self):
        return self._usedChars

    def build(self, v):
        self.clear()
        size = len(v)
        if isinstance(v, str):
            # convert to JavaScript compatible string
            rawstring = v.encode('utf_16_le')
            v = struct.unpack("<%dH" % size, rawstring)
        self._usedChars = set(v)
        bitsize = self.bitsize()
        for i in _range(bitsize):
            self._bv.append(bitvector.BitVector())
            self._seps.append(0)
        self._size = size
        for i, c in enumerate(v):
            self._bv[0].set(i, self._uint2bit(c, 0, bitsize))
        self._bv[0].build()
        self._seps[0] = self._bv[0].size0()
        self._range[0] = 0
        self._range[1] = self._seps[0]

        depth = 1
        while depth < bitsize:
            range_tmp = copy.copy(self._range)
            for i, code in enumerate(v):
                bit = self._uint2bit(code, depth, bitsize)
                key = code >> (bitsize - depth)
                self._bv[depth].set(range_tmp[key], bit)
                range_tmp[key] += 1
            self._bv[depth].build()
            self._seps[depth] = self._bv[depth].size0()

            range_rev = {}
            for range_key, value in self._range.items():
                if value != range_tmp[range_key]:
                    range_rev[value] = range_key
            self._range = {}
            pos0 = 0
            pos1 = self._seps[depth]
            for begin, value in sorted(range_rev.items()):
                end = range_tmp[value]
                num0  = self._bv[depth].rank(end, False) - self._bv[depth].rank(begin, False)
                num1  = end - begin - num0
                if num0 > 0:
                    self._range[value << 1] = pos0
                    pos0 += num0
                if num1 > 0:
                    self._range[(value << 1) + 1] = pos1
                    pos1 += num1
            depth += 1

    def size(self):
        return self._size
    
    def count(self, c):
        return self.rank(self.size(), c)
    
    def get(self, i):
        if i >= self.size():
            raise RangeError("WaveletMatrix.get() : range error")
        value = 0
        depth = 0
        bitsize = self.bitsize()
        while depth < bitsize:
            bit = self._bv[depth].get(i)
            i = self._bv[depth].rank(i, bit)
            value <<= 1
            if bit:
                i += self._seps[depth]
                value += 1
            depth += 1
        return value

    def rank(self, i, c):
        if i > self.size():
            raise RangeError("WaveletMatrix.rank(): range error")
        if i == 0:
            return 0
        begin = self._range.get(c)
        if begin == None:
            return 0
        end   = i
        depth = 0
        bitsize = self.bitsize()
        while depth < bitsize:
            bit = self._uint2bit(c, depth, bitsize)
            end = self._bv[depth].rank(end, bit)
            if bit:
                end += self._seps[depth]
            depth += 1
        return end - begin
    
    def rank_less_than(self, i, c):
        if i > self.size():
            raise RangeError("WaveletMatrix.rank_less_than(): range error")
        if i == 0:
            return 0
        begin = 0
        end   = i
        depth = 0
        rlt   = 0
        bitsize = self.bitsize()
        while depth < bitsize:
            rank0_begin = self._bv[depth].rank(begin, False)
            rank0_end   = self._bv[depth].rank(end,   False)
            if self._uint2bit(c, depth, bitsize):
                rlt += (rank0_end - rank0_begin)
                begin += (self._seps[depth] - rank0_begin)
                end   += (self._seps[depth] - rank0_end)
            else:
                begin = rank0_begin
                end   = rank0_end
            depth += 1
        return rlt
    
    def dump(self, output):
        output.dump_16bit_number(self._maxcharcode)
        output.dump_16bit_number(self.bitsize())
        output.dump_32bit_number(self._size)
        for i in _range(self.bitsize()):
            self._bv[i].dump(output)
        for i in _range(self.bitsize()):
            output.dump_32bit_number(self._seps[i])
        output.dump_32bit_number(len(self._range))
        for key, value in self._range.items():
            output.dump_32bit_number(key)
            output.dump_32bit_number(value)

    def load(self, input):
        self.clear()
        self._maxcharcode = input.load_16bit_number()
        self._bitsize = input.load_16bit_number()
        self._size = input.load_32bit_number()
        for i in _range(self.bitsize()):
            bit_vector = bitvector.BitVector()
            bit_vector.load(input)
            self._bv.append(bit_vector)
        for i in _range(self.bitsize()):
            self._seps.append(input.load_32bit_number())
        range_size = input.load_32bit_number()
        for i in _range(range_size):
            key = input.load_32bit_number()
            value = input.load_32bit_number()
            self._range[key] = value

    def _uint2bit(self, c, i, bitsize):
        return ((c >> (bitsize - 1 - i)) & 0x1) == 0x1




'''
self is a Python version of shellinford library:
https://code.google.com/p/shellinford/

License: http://shibu.mit-license.org/
'''

from . import binaryio
import math
import array

_range = getattr(__builtins__, 'xrange', range)

class BitVector(object):
    SMALL_BLOCK_SIZE =  32
    LARGE_BLOCK_SIZE = 256
    BLOCK_RATE       =   8

    def __init__(self):
        self._r = array.array('L')
        self._v = array.array('L')
        self.clear()

    def build(self):
        self._size1 = 0
        for i, value in enumerate(self._v):    
            if i % BitVector.BLOCK_RATE == 0:
                self._r.append(self.size1())
            self._size1 += self._rank32(self._v[i], BitVector.SMALL_BLOCK_SIZE)

    def clear(self):
        del self._v[:]
        del self._r[:]
        self._size = 0
        self._size1 = 0

    def size(self):
        return self._size

    def size0(self):
        '''
        :return: size
        :rtype: int
        '''
        return self._size - self._size1

    def size1(self):
        '''
        :return: size
        :rtype: int
        '''
        return self._size1

    def set(self, value, flag = True):
        '''
        :param value: value
        :type  value: int
        :param flag: flag
        :type  flag: bool
        '''
        if value >= self.size():
            self._size = value + 1
        
        q = value // BitVector.SMALL_BLOCK_SIZE
        r = value % BitVector.SMALL_BLOCK_SIZE
        while q >= len(self._v):
            self._v.append(0)
        m = 0x1 << r
        if flag:
            self._v[q] |=  m
        else:
            self._v[q] &= ~m

    def get(self, value):
        '''
        :param value: value
        :type  value: int
        :return: bit
        :rtype: bool
        '''
        if value >= self.size():
            raise RangeError("BitVector.get() : range error")
        q = value // BitVector.SMALL_BLOCK_SIZE
        r = value % BitVector.SMALL_BLOCK_SIZE
        m  = 0x1 << r
        return bool(self._v[q] & m)

    def rank(self, i, b = True):
        '''
        :param i: position
        :type  i: int
        :param b: invert
        :type  b: bool
        :return: rank value
        :rtype: int
        '''
        if i > self.size():
            raise RangeError("BitVector.rank() : range error")
        if i == 0:
            return 0
        i -= 1
        q_large = int(math.floor(i // BitVector.LARGE_BLOCK_SIZE))
        q_small = int(math.floor(i // BitVector.SMALL_BLOCK_SIZE))
        r       = int(math.floor(i % BitVector.SMALL_BLOCK_SIZE))
        rank = self._r[q_large]
        if not b:
            rank = q_large * BitVector.LARGE_BLOCK_SIZE - rank
        begin = q_large * BitVector.BLOCK_RATE
        for j in _range(begin, q_small):
            if b:
                value = self._v[j]
            else:
                value = ~self._v[j]
            rank += self._rank32(value, BitVector.SMALL_BLOCK_SIZE)
        if b:
            value = self._v[q_small]
        else:
            value = ~self._v[q_small]
        return rank + self._rank32(value, r + 1)

    def select(self, i, b = True):
        '''
        :param i: i
        :type  i: int
        :param b: b
        :type  b: bool
        :return: result
        :rtype: int
        '''
        if b:
            if i >= self.size1():
                raise RangeError("BitVector.select() : range error")
        elif i >= self.size0(): 
            raise RangeError("BitVector.select() : range error")
        left = 0
        right = len(self._r)
        while left < right:
            pivot = int(math.floor((left + right) // 2))
            rank  = self._r[pivot]
            if not b:
                rank = pivot * BitVector.LARGE_BLOCK_SIZE - rank
            if i < rank:
                right = pivot
            else:
                left = pivot + 1
        right -= 1
        if b:
            i -= self._r[right]
        else:
            i -= right * BitVector.LARGE_BLOCK_SIZE - self._r[right]
        j = right * BitVector.BLOCK_RATE
        while True:
            if b:
                value = self._v[j]
            else:
                value = ~self._v[j]
            rank = self._rank32(value, BitVector.SMALL_BLOCK_SIZE)
            if i < rank:
                break
            j += 1
            i -= rank
        return j * BitVector.SMALL_BLOCK_SIZE + self._select32(self._v[j], i, b)

    def _rank32(self, x, i):
        '''
        :param x: x
        :type  x: int
        :param i: i
        :type  i: int
        :param b: b
        :type  b: bool
        '''
        x <<= (BitVector.SMALL_BLOCK_SIZE - i)
        x = ((x & 0xaaaaaaaa) >>  1) + (x & 0x55555555)
        x = ((x & 0xcccccccc) >>  2) + (x & 0x33333333)
        x = ((x & 0xf0f0f0f0) >>  4) + (x & 0x0f0f0f0f)
        x = ((x & 0xff00ff00) >>  8) + (x & 0x00ff00ff)
        x = ((x & 0xffff0000) >> 16) + (x & 0x0000ffff)
        return x
    

    def _select32(self, x, i, b):
        '''
        :param x: x
        :type  x: int
        :param i: i
        :type  i: int
        :param b: b
        :type  b: bool
        '''
        if not b:
            x = ~x
        x1 = ((x  & 0xaaaaaaaa) >>  1) + (x  & 0x55555555)
        x2 = ((x1 & 0xcccccccc) >>  2) + (x1 & 0x33333333)
        x3 = ((x2 & 0xf0f0f0f0) >>  4) + (x2 & 0x0f0f0f0f)
        x4 = ((x3 & 0xff00ff00) >>  8) + (x3 & 0x00ff00ff)
        x5 = ((x4 & 0xffff0000) >> 16) + (x4 & 0x0000ffff)
        i += 1
        pos = 0
        v5 = x5 & 0xffffffff
        if i > v5:
            i -= v5
            pos += 32
        v4 = (x4 >> pos) & 0x0000ffff
        if i > v4:
            i -= v4
            pos += 16
        v3 = (x3 >> pos) & 0x000000ff
        if i > v3:
            i -= v3
            pos += 8
        v2 = (x2 >> pos) & 0x0000000f
        if i > v2:
            i -= v2
            pos += 4
        v1 = (x1 >> pos) & 0x00000003
        if i > v1:
            i -= v1
            pos += 2
        v0 = (x >> pos) & 0x00000001
        if i > v0:
            i -= v0
            pos += 1
        return pos

    def dump(self, output):
        '''
        :param output: output
        :type  output: BinaryOutput
        '''
        output.dump_32bit_number(self._size)
        output.dump_32bit_number_list(self._v.tolist())

    def load(self, input):
        '''
        :param input: input
        :type  input: BinaryInput
        '''
        self.clear()
        self._size = input.load_32bit_number()
        self._v.fromlist(input.load_32bit_number_list())
        self.build()



from . import bitvector
from . import binaryio


class Metadata(object):
    def __init__(self, parent, name):
        self._parent = parent
        self._bitVector = bitvector.BitVector()
        self._name = name

    def _size(self):
        return self._bitVector.rank(self._bitVector.size(), 1)

    def get_content(self, index):
        '''
        :param index: index
        :type  index: int
        '''
        if index < 0 or self._size() <= index:
            raise IndexError("Section.getContent() : range error %d" % index)
        startPosition = 0
        if index > 0:
            startPosition = self._bitVector.select(index - 1, 1) + 1
        length = self._bitVector.select(index, 1) - startPosition + 1
        return self._parent._get_substring(startPosition, length)

    def get_content_with_EOB(self, index):
        '''
        :param index: index
        :type  index: int
        '''
        if index < 0 or self._size() <= index:
            raise IndexError("Section.getContent() : range error %d" % index)
        
        startPosition = 0
        if index > 0:
            startPosition = self._bitVector.select(index - 1, 1) + 1
        
        length = self._bitVector.select(index, 1) - startPosition + 1
        return self._parent._get_substring_with_EOB(startPosition, length)

    def get_start_position(self, index):
        '''
        :param index: index
        :type  index: int
        '''
        if index < 0 or self._size() <= index:
            raise IndexError("Section.getContent() : range error %d" % index)
        
        startPosition = 0
        if index > 0:
            startPosition = self._bitVector.select(index - 1, 1) + 1
        
        return startPosition
    
    def grouping(self, result, positions, word, stemmed):
        '''
        :param result: result
        :type  result: SingleResult
        :param positions: positions
        :type  positions: int[]
        :param word: word
        :type  word: str
        :param stemmed: stemmed
        :type  stemmed: bool
        '''
        pass

    def get_information(self, index):
        '''
        :param index: index
        :type  index: int
        '''
        pass

    def _build(self):
        self._bitVector.build()

    def _load_parent(self, input):
        self._bitVector.load(input)
        self._parent._metadataLabels.append(self._name)
        self._parent._metadatas[self._name] = self

    def _dump(self, output):
        output.dump_string(self._name)
        self._bitVector.dump(output)
    

class Section(Metadata):
    TypeID = 0

    def __init__(self, parent, name):
        super(Section, self).__init__(parent, name)
        self._names = []
    
    def set_tail(self, name, index=None):
        '''
        :param name: name
        :type  name: str
        :param index: index
        :type  index: int
        '''
        if index is None:
            index = self._parent.content_size() - 1
        if self._parent._isLastEob:
            raise ValueError("Tail should not be 'eof' or 'eob'")
        self._names.append(name)
        self._bitVector.set(index, 1)

    def size(self):
        '''
        :return: section count
        :rtype: int
        '''
        return len(self._names)

    def get_section_index(self, position):
        '''
        :param position: position
        :type  position: int
        :return:
        :rtype: int
        '''
        if position < 0 or self._bitVector.size() <= position:
            raise IndexError("Section.getSectionIndex() : range error %d" % position)
        return self._bitVector.rank(position, 1)

    def get_name(self, index):
        '''
        :param index: index
        :type  index: int
        :return: name
        :rtype: str
        '''
        if index < 0 or self.size() <= index:
            raise IndexError("Section.getName() : range error")
        return self._names[index]

    def grouping(self, result, positions, word, stemmed):
        '''
        :param result: result
        :type  result: SingleResult
        :param positions: positions
        :type  positions: int[]
        :param word: word
        :type  word: str
        :param stemmed: stemmed
        :type  stemmed: bool
        '''
        for position in positions:
            index = self.getSectionIndex(position)
            unit = result.getSearchUnit(index)
            if unit.startPosition < 0:
                unit.startPosition = self.getStartPosition(index)
            unit.addPosition(word, position - unit.startPosition, stemmed)

    def get_information(self, index):
        '''
        :param index: index
        :type  index: int
        '''
        return self.getName(index)

    @staticmethod
    def _load(parent, input):
        name = input.load_string()
        section = Section(parent, name)
        section._load_parent(input)
        section._names = input.load_string_list()

    def _dump(self, output):
        output.dump_16bit_number(Section.TypeID)
        super(Section, self)._dump(output)
        output.dump_string_list(self._names)
    

class Splitter(Metadata):
    TypeID = 1

    def __init__(self, parent, name):
        super(Splitter, self).__init__(parent, name)

    def size(self):
        '''
        :return: section count
        :rtype: int
        '''
        return self._size()

    def split(self, index = None):
        '''
        :param index: index
        :type  index: int
        '''
        if index is None:
            index = self._parent.content_size() - 1
        if self._parent._isLastEob:
            raise ValueError("Tail should not be 'eof' or 'eob'")
        self._bitVector.set(index, 1)

    def get_index(self, position):
        '''
        :param position: position
        :type  position: int
        '''
        if position < 0 or self._bitVector.size() <= position:
            raise IndexError("Section.getSectionIndex() : range error")
        return self._bitVector.rank(position, 1)

    def grouping(self, result, positions, word, stemmed):
        '''
        :param result: result
        :type  result: SingleResult
        :param positions: positions
        :type  positions: int[]
        :param word: word
        :type  word: str
        :param stemmed: stemmed
        :type  stemmed: bool
        '''
        for position in positions:
            index = self.get_index(position)
            unit = result.get_search_unit(index)
            if unit.startPosition < 0:
                unit.startPosition = self.getStartPosition(index)
            unit.addPosition(word, position - unit.startPosition, stemmed)

    def get_information(self, index):
        '''
        :param index: index
        :type  index: int
        '''
        return "%s %d" % (self._name, index + 1)

    @staticmethod
    def _load(parent, input):
        name = input.load_string()
        section = Splitter(parent, name)
        section._load_parent(input)

    def _dump(self, output):
        output.dump_16bit_number(Splitter.TypeID)
        super(Splitter, self)._dump(output)


class Table(Metadata):
    TypeID = 2

    def __init__(self, parent, name, headers = None):
        super(Table, self).__init__(parent, name)
        if headers is not None:
            self._headers = headers
        self._columnTails = bitvector.BitVector()

    def row_size(self):
        return self._size()

    def column_size(self):
        return len(self._headers)

    def set_column_tail(self):
        index = self._parent.content_size()
        if self._parent._isLastEob:
            raise Error("Tail should not be 'eof' or 'eob'")
        self._columnTails.set(index, 1)

    def set_column_tail_and_EOB(self):
        self.set_column_tail()
        self._parent.add_end_of_block()

    def set_row_tail(self):
        index = self._parent.content_size() - 1
        self._bitVector.set(index, 1)
    
    def get_cell(self, position):
        if position < 0 or self._bitVector.size() <= position:
            raise IndexError("Section.getSectionIndex() : range error %d" % position)
        row = self._bitVector.rank(position, 1)
        currentColumn = self._columnTails.rank(position, 1)
        lastRowColumn = 0
        if row > 0:
            startPosition = self._bitVector.select(row - 1, 1) + 1
            lastRowColumn = self._columnTails.rank(startPosition, 1)
        result = [row, currentColumn - lastRowColumn]
        return result

    def get_row_content(self, rowIndex):
        content = self.get_content_with_EOB(rowIndex)
        values = content.split("\x01", len(self._headers))
        result = {}
        for i, header in enumerate(self._headers):
            if i < len(values):
                result[header] = values[i]
            else:
                result[header] = ''
        return result

    def grouping(self, result, positions, word, stemmed):
        '''
        :param result: result
        :type  result: SingleResult
        :param positions: positions
        :type  positions: int[]
        :param word: word
        :type  word: str
        :param stemmed: stemmed
        :type  stemmed: bool
        '''
        # TODO implement
        pass

    def get_information(self, index):
        '''
        :param index: index
        :type  index: int
        '''
        return ''

    def _build(self):
        self._bitVector.build()
        self._columnTails.build()

    @staticmethod
    def _load(parent, input):
        name = input.load_string()
        table = Table(parent, name)
        table._load_parent(input)
        table._headers = input.load_string_list()
        table._columnTails.load(input)

    def _dump(self, output):
        output.dump_16bit_number(Table.TypeID)
        super(Table, self)._dump(output)
        output.dump_string_list(self._headers)
        self._columnTails.dump(output)

class Block(Metadata):
    TypeID = 3

    def __init__(self, parent, name):
        super(Block, self).__init__(parent, name)
        self._names = []
        self._start = False

    def start_block(self, blockName, index = None):
        if index is None:
            index = self._parent.content_size() - 1
        if self._start:
            raise ValueError('Splitter `%s` is not closed' % self._names[-1])
        self._start = True
        self._names.append(blockName)
        self._bitVector.set(index, 1)

    def end_block(self, index=None):
        if index is None:
            index = self._parent.content_size() - 1
        if self._parent._isLastEob:
            raise ValueError("Block end should not be 'eof' or 'eob'")
        if not self._start:
            raise ValueError('Splitter is not started')
        self._start = False
        self._bitVector.set(index, 1)

    def size(self):
        '''
        :return: section count
        :rtype: int
        '''
        return len(self._names)

    def block_index(self, position):
        if position < 0 or (self._parent._fmindex.size() - 1) <= position:
            raise IndexError("Block.block_index() : range error %d" % position)
        if position >= self._bitVector.size():
            position = self._bitVector.size() - 1
            result = self._bitVector.rank(position, 1) + 1
        else:
            result = self._bitVector.rank(position, 1)
        return result

    def in_block(self, position):
        block_index = self.block_index(position)
        return (block_index % 2) != 0

    def get_block_content(self, position):
        block_index = self.block_index(position)
        if block_index % 2 != 0:
            result = self.get_content(block_index)
        else:
            result = ''
        return result

    def get_block_name(self, position):
        block_index = self.block_index(position)
        if block_index % 2 != 0:
            result = self._names[block_index >> 1]
        else:
            result = ''
        return result

    def get_information(self, index):
        '''
        :param index: index
        :type  index: int
        '''
        return ''

    @staticmethod
    def _load(parent, input):
        name = input.load_string()
        block = Block(parent, name)
        block._load_parent(input)
        block._names = input.load_string_list()

    def _dump(self, output):
        output.dump_16bit_number(Block.TypeID)
        super(Block, self)._dump(output)
        output.dump_string_list(self._names)



from . sais import SAIS

_range = getattr(__builtins__, 'xrange', range)

class BWT(object):

    END_MARKER = chr(0)
    END_MARKER_CODE = 0

    def __init__(self, string, rawmode=False):
        self._rawmode = rawmode
        if rawmode:
            str2 = string + [0];
            self._str = str2;
            self._size = len(self._str)
            self._suffixarray = SAIS.make(str2, rawmode=True);
            self._head = self._suffixarray.index(0);
        else:
            str2 = string + BWT.END_MARKER;
            self._str = str2;
            self._size = len(self._str)
            self._suffixarray = SAIS.make(str2);
            self._head = self._suffixarray.index(0);

    def size(self):
        return self._size;

    def head(self):
        return self._head;

    def clear(self):
        self._str = "";
        self._size = 0;
        self._head = 0;
        del self._suffixarray[:]

    def get(self, i = None, replace = None):
        if i is None:
            size = self.size();
            chars = [self.get(i) for i in _range(size)]
            if self._rawmode:
                result = chars
                if replace is not None:
                    while BWT.END_MARKER_CODE in replace:
                        index = result.index(BWT.END_MARKER_CODE)
                        result[index] = replace
            else:
                result = ''.join(chars)
                if replace is not None:    
                    result = result.replace(BWT.END_MARKER, replace);
        else:
            size = self.size();
            if i >= size:
                raise RangeError("BWT.get() : range error");
            index = (self._suffixarray[i] + size - 1) % size;
            result = self._str[index];
        return result

def bwt(string, endMarker=None):
    bwt = BWT(string);
    return bwt.get(replace=endMarker)



from setuptools import setup, find_packages
import sys, os

version = '0.2.0'

setup(name='oktavia',
      version=version,
      description="High performance pure Python/JavaScript search engine",
      long_description="""This is search engine is pure Python/JavaScript on-memory search engine.

* It uses FM-Index as an algorithm.
* It is good for eastern asian languages (CJK) because it doesn't use an invert index.
* It works on completely on memory.
* There are JavaScript index generator/search engine too. You can provide search feature to you web applications and static HTML.

""",
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Environment :: Web Environment',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: JavaScript',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Topic :: Internet :: WWW/HTTP :: Indexing/Search'
      ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='search python javascript fmindex',
      author='Yoshiki Shibukawa',
      author_email='yoshiki@shibu.jp',
      url='https://github.com/shibukawa/oktavia.py/',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=True,
      install_requires=[
          # -*- Extra requirements: -*-
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



class Query(object):
    def __init__(self):
        self.word = ''
        self.OR = False
        self.NOT = False
        self.RAW = False

    def __str__(self):
        result = []
        if self.OR:
            result.push("OR ")
        if self.NOT:
            result.push("-")
        if self.RAW:
            result.push('"', self.word, '"')
        else:
            result.push(self.word)
        return ''.join(result)



'''
This is a Python version of shellinford library:
https:#code.google.com/p/shellinford/

License: http:#shibu.mit-license.org/
'''

import math
import struct

from . import waveletmatrix
from . import bwt
from . import binaryio

try:
    _range = xrange
except NameError:
    _range = range

try:
    _strtype = basestring
except NameError:
    _strtype = str

class FMIndex(object):
    def __init__(self, rawmode = False):
        self._ddic = 0
        self._head = 0
        self._substr = []
        self._sv = waveletmatrix.WaveletMatrix()
        self._posdic = []
        self._idic = []
        self._rlt = [0] * 65536
        self._build = False
        self._rawmode = rawmode

    def clear(self):
        self._sv.clear()
        del self._posdic[:]
        del self._idic[:]
        self._ddic = 0
        self._head = 0
        self._rlt = [0] * 65536
        del self._substr[:]

    def size(self):
        return self._sv.size()
    
    def content_size(self):
        return sum((len(s) for s in self._substr))

    def rank(self, pos, code):
        return self._sv.rank(pos, code)

    def get(self, pos):
        return self._sv.get(pos)

    def get_rows(self, key, pos=None):
        import sys
        if isinstance(key, _strtype):
            # convert to JavaScript compatible string
            rawstring = key.encode('utf_16_le')
            key = struct.unpack("<%dH" % len(key), rawstring)
        i = len(key) - 1
        c = key[i]
        first = self._rlt[c] + 1
        last  = self._rlt[c + 1]
        while first <= last:
            if i == 0:
                if pos is not None:
                    first -= 1
                    last -= 1
                    del pos[:]
                    pos.append(first)
                    pos.append(last)
                return (last - first  + 1)
            i -= 1
            c = key[i]
            first = self._rlt[c] + self._sv.rank(first - 1, c) + 1
            last  = self._rlt[c] + self._sv.rank(last,      c)
        return 0

    def get_position(self, i):
        if i >= self.size():
            raise RangeError("FMIndex.get_position() : range error")
        pos = 0
        while i != self._head:
            if (i % self._ddic) == 0:
                pos += (self._posdic[i // self._ddic] + 1)
                break
            c = self._sv.get(i)
            i = self._rlt[c] + self._sv.rank(i, c) #LF
            pos += 1
        return pos % self.size()

    def _join(self):
        if self._rawmode:
            doc = []
            for substr in self._substr:
                doc = doc + substr
        else:
            doc = "".join(self._substr)
        return doc

    def get_substring(self, pos, length):
        if not self._build:
            doc = self._join()
            return doc[pos : pos + length]

        if pos >= self.size():
            raise RangeError("FMIndex.get_substring() : range error")
        pos_end = min(pos + length, self.size())
        pos_tmp = self.size() - 1
        i = self._head
        pos_idic = (pos_end + self._ddic - 2) // self._ddic
        if pos_idic < len(self._idic):
            pos_tmp = pos_idic * self._ddic
            i = self._idic[pos_idic]
        codes = []
        while pos_tmp >= pos:
            c = self.get(i)
            i = self._rlt[c] + self.rank(i, c) #LF
            if pos_tmp < pos_end and c != 0:
                codes.insert(0, c)
            if pos_tmp == 0:
                break
            pos_tmp -= 1

        if self._rawmode:
            return codes
        return struct.pack('<%dH' % len(codes), *codes).decode('utf_16_le')

    def build(self, ddic, maxChar=65535):
        doc = self._join()
        sa = bwt.BWT(doc, rawmode=self._rawmode)
        s = sa.get()
        self._ssize = len(s)
        self._head = sa.head()
        del self._substr[:]
        self._sv.set_max_char_code(maxChar)
        self._sv.build(s)
        size = self.size()
        for c in _range(maxChar):
            self._rlt[c] = self._sv.rank_less_than(size, c)
        self._rlt[maxChar] = 0;
        self._ddic = int(ddic)
        self._buildDictionaries()
        self._build = True

    def _buildDictionaries(self):
        del self._posdic[:]
        del self._idic[:]
        for i in _range(self._ssize // self._ddic + 1):
            self._posdic.append(0)
            self._idic.append(0)
        i = self._head
        pos = self.size() - 1

        if (i % self._ddic) == 0:
            self._posdic[i // self._ddic] = pos
        if (pos % self._ddic) == 0:
            self._idic[pos // self._ddic] = i
        c = self._sv.get(i)
        i = self._rlt[c] + self._sv.rank(i, c) #LF
        pos -= 1
        while i != self._head:
            if (i % self._ddic) == 0:
                self._posdic[i // self._ddic] = pos
            if (pos % self._ddic) == 0:
                self._idic[pos // self._ddic] = i
            c = self._sv.get(i)
            i = self._rlt[c] + self._sv.rank(i, c) #LF
            pos -= 1

    def append(self, doc):
        if len(doc) == 0:
            raise ValueError("FMIndex::append(): empty string")
        self._substr.append(doc)

    def search(self, keyword):
        result = []
        position = []
        rows = self.get_rows(keyword, position)
        if rows > 0:
            first = position[0]
            last = position[1]
            for i in _range(first, last + 1):
                result.append(self.get_position(i))
        return result

    def dump(self, output):
        output.dump_32bit_number(self._ddic)
        output.dump_32bit_number(self._ssize)
        output.dump_32bit_number(self._head)
        self._sv.dump(output)
        output.dump_32bit_number(len(self._posdic))
        for v in self._posdic:
            output.dump_32bit_number(v)
        for v in self._idic:
            output.dump_32bit_number(v)

    def load(self, input):
        self.clear()
        self._ddic = int(input.load_32bit_number())
        self._ssize = input.load_32bit_number()
        self._head = input.load_32bit_number()
        self._sv.load(input)
        maxChar = self._sv.max_char_code()
        size = self._sv.size()
        self._rlt = []
        for c in _range(maxChar):
            self._rlt.append(self._sv.rank_less_than(size, c))
        self._rlt.append(0)
        size = input.load_32bit_number()
        del self._posdic[:]
        del self._idic[:]
        for i in _range(size):
            self._posdic.append(input.load_32bit_number())
        for i in _range(size):
            self._idic.append(input.load_32bit_number())
        self._build = True

